from django.template.loaders import filesystem as fs
from django.template import TemplateDoesNotExist, Template as DjTemplate
from django.template.utils import get_app_template_dirs


class TemplateAppy(DjTemplate):
    def __init__(self, template_string, origin=None, name=None, engine=None):
        if engine is None:
            from .engine import Engine
            engine = Engine.get_default()
        if origin is None:
            origin = Origin(UNKNOWN_SOURCE)
        self.name = name
        self.origin = origin
        self.engine = engine
        self.source = str(template_string)  # May be lazy.
        # self.nodelist = self.compile_nodelist()


class FilesystemLoader(fs.Loader):

    def get_contents(self, origin):
        # ODT is a compressed format not a string
        # we need to read the template w/o interpreting as byte stream
        try:
            with open(origin.name, 'rb') as fp:
                return fp.read()
        except FileNotFoundError:
            raise TemplateDoesNotExist(origin)

    def get_template(self, template_name, skip=None, *args, **kwargs):
        """
        Call self.get_template_sources() and return a Template object for
        the first template matching template_name. If skip is provided, ignore
        template origins in skip. This is used to avoid recursion during
        template extending.
        """
        tried = []

        for origin in self.get_template_sources(template_name):
            if skip is not None and origin in skip:
                tried.append((origin, 'Skipped'))
                continue

            try:
                contents = self.get_contents(origin)
            except TemplateDoesNotExist:
                tried.append((origin, 'Source does not exist'))
                continue
            else:
                return TemplateAppy(
                    contents, origin, origin.template_name, self.engine,
                )

        raise TemplateDoesNotExist(template_name, tried=tried)

    def get_template_sources(self, template_name):
        """
        Return an Origin object pointing to an absolute path in each directory
        in template_dirs. For security reasons, if a path doesn't lie inside
        one of the template_dirs it is excluded from the result set.
        """
        # no point in searching if this is the wrong loader
        # this way we can set AppyPodEngine in first position in TEMPLATES
        if not (template_name.endswith('.odt') or template_name.endswith('.ods')):
            raise TemplateDoesNotExist(template_name)

        for template_dir in self.get_dirs():
            try:
                name = fs.safe_join(template_dir, template_name)
            except fs.SuspiciousFileOperation:
                # The joined path was located outside of this template_dir
                # (it might be inside another one, so this isn't fatal).
                continue
            yield fs.Origin(
                name=name,
                template_name=template_name,
                loader=self,
            )


class AppDirectoriesLoader(FilesystemLoader):

    def get_dirs(self):
        return get_app_template_dirs('templates')
